#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 13 07:50:55 2019

@author: fuaa
"""
from matplotlib import pyplot as plt
import numpy as np
import getpass
import socket
#import cv2
from matplotlib.colors import LinearSegmentedColormap

_THRESHOLD = 10e-9


#########################################################################
def validar_resultado(*args, **kwargs):
    """"
    Función para validar resultado a invocar desde el notebook.
    """
    # _ES_CORRECTO = False
    _DEBUG = False

    # Abrir el cartel del validación.
    print("+-------------------------------------------------------------------------+")
    print("|               FuAA (2do. parcial 2022): validar resultado               |")
    print("+-------------------------------------------------------------------------+")
    for key, value in kwargs.items():
        if key == "debug":
            _DEBUG = value
    if _DEBUG:
        print('args:', args)
        print('kwargs:', kwargs)

    if (len(args) == 0):
        print("| Sin opciones para evaluar.                                              |")
        print("+-------------------------------------------------------------------------+")
        return False

    ###########################################################
    # Sección I. Parte (a)
    ###########################################################
    elif (args[0] == "preparar_datos_ej_svm_fuaa"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return
        
        x_test = np.vstack((np.ones((5,28,28)), np.zeros((5,28,28))))

        x_pre, scaler_ = test_algoritmo(x_test)
        
        x_pre_true = np.vstack((np.ones((5,784)), (-1)*np.ones((5,784))))

        mean_true = (0.5) * np.ones(784)
        std_true = (0.5) * np.ones(784)

        # Validar 
        condicion1 = x_pre.shape == x_pre_true.shape
        fuaa_assert(condicion1,
            mensajeFalse="Revise las dimensiones de la matriz de salida.",
            mensajeTrue= "Dimensiones correctas.")

        # Validar 
        condicion2 = son_iguales(scaler_.mean_ , mean_true) and son_iguales(scaler_.scale_, std_true) 
        fuaa_assert(condicion2,
            mensajeFalse="Revise la estandarización.",
            mensajeTrue= "Estandarización correcta.")


        condicion3 = son_iguales(x_pre, x_pre_true) 
        fuaa_assert(condicion3,
                    mensajeFalse="Revise la implementación.",
                    mensajeTrue= "Resultado válido")


    ###########################################################
    # Ejercicio Sección II. Parte (b)
    ###########################################################
    elif (args[0] == "obtener_desempeno"):
        if 'funcion' in kwargs:
            test_algoritmo = kwargs['funcion']
        else:
            print( "[validar_resultado] Error: llamada sin argumentos (%s)" % args[0] )
            return
        
        y_true_ex = np.array([1,0,0,0,1,1,0,1]) 
        y_pred_ex = np.array([1,0,1,0,1,1,0,0])

        error, accuracy = test_algoritmo(y_true_ex, y_pred_ex)

        
        error_true = 0.25
        accuracy_true = 75.0

        accuracy_no_porcentual_true = 0.75

        # Validar error
        condicion1 = son_iguales(error, error_true) 
        fuaa_assert(condicion1,
                    mensajeFalse="El cálculo del error NO es correcto.",
                    mensajeTrue="Cálculo del error correcto.")

        # Validar accuracy.
        condicion2 = son_iguales(accuracy, accuracy_true)  

        fuaa_assert(condicion2,
                    mensajeFalse="El cálculo del accuracy NO es correcto.",
                    mensajeTrue= "Cálculo del accuracy correcto.")

        # Validar accuracy.
        condicion3 = not (son_iguales(accuracy, accuracy_no_porcentual_true))  

        fuaa_assert(condicion3,
                    mensajeFalse="El accuracy no esta en valor porcentual.",)
        
   
    ###########################################################
    # No hay ninguna opción de ejercicio.
    ###########################################################
    else:
        printcolor("Ninguna opción revisada.")

    # Cerrar el cartel.
    print("+-------------------------------------------------------------------------+")


#########################################################################
def identificar_parcial():
    username = getpass.getuser()
    hostname = socket.gethostname()
    print("Usuario %s en %s." % (username, hostname))


#########################################################################
def printcolor(mensaje, color="k"):
    """
    Función para imprimir en colores y con el formato de interés.
    """
    if (color == "r"):
        mensajeColor = "\x1b[31m" + mensaje + "\x1b[0m"
    elif (color == "g"):
        mensajeColor = "\x1b[32m" + mensaje + "\x1b[0m"
    elif (color == "y"):
        mensajeColor = "\x1b[33m" + mensaje + "\x1b[0m"
    elif (color == "b"):
        mensajeColor = "\x1b[34m" + mensaje + "\x1b[0m"
    elif (color == "p"):
        mensajeColor = "\x1b[35m" + mensaje + "\x1b[0m"
    elif (color == "c"):
        mensajeColor = "\x1b[36m" + mensaje + "\x1b[0m"
    else:
        mensajeColor = mensaje
    mensaje_out = " " + mensajeColor
    print(mensaje_out)


#########################################################################
def fuaa_assert(condicion,
                mensajeFalse="El resultado no es válido.",
                mensajeTrue="Resultado validado."):
    """
    Función similar al assert.
    """
    # Custom assert.
    if (condicion):
        printcolor(mensajeTrue, "g")
    else:
        printcolor(mensajeFalse, "r")


#########################################################################
def son_iguales(x1, x2, threshold=_THRESHOLD):
    """
    Evaluar si dos elementos son iguales o no, con una tolerancia dada (threshold).
    """
    if isinstance(x1, float) or isinstance(x1, int):
        dif = np.abs(x1 - x2)
        condicion = (dif < threshold)
    elif isinstance(x1, np.ndarray):
        dif = np.sqrt(np.sum((x1 - x2)**2)) / x1.size
        condicion = (dif < threshold)
    else:
        printcolor("Ningún tipo validado para son_iguales()", "r")
        condicion = False
    return condicion


#########################################################################
def validar_parametros(parametros, min_params):
    """
    Validar número de parámetros.
    """
    condicion = (len(parametros) >= min_params)
    if not condicion:
        printcolor("[validar_resultado] Insuficientes parámetros (\"%s\"), se necesitan %d, hay %d." %
                   (parametros[0], min_params, len(parametros)), "r")
    return condicion


#########################################################################
def load_img(file, scale_percent=70):

    img = cv2.imread(file)

    scale_percent = 70 # percent of original size
    width = int(img.shape[1] * scale_percent / 100)
    height = int(img.shape[0] * scale_percent / 100)
    dim = (width, height)

    # resize image
    resized = cv2.resize(img, dim, interpolation = cv2.INTER_AREA)
    img = np.asarray(cv2.cvtColor(resized, cv2.COLOR_BGR2RGB), dtype=np.float64) / 255.
    
    return img
#########################################################################
def plot_colorspace(img, centros=[], elev=35, azim=-35):
    fig = plt.figure(figsize=(16,16))
    axis = fig.add_subplot(1, 1, 1, projection="3d",computed_zorder=False) # 3D plot with scalar values in each axis

    im = img[:,:]
    r, g, b = list(im[:,:,0].flatten()), list(im[:,:,1].flatten()), list(im[:,:,2].flatten())

    axis.scatter(r, g, b, facecolor=im.reshape(-1,3), marker="o",edgecolors='k',linewidth=0.02,alpha=0.3,zorder=2)
    for clust in range(len(centros)):
        axis.scatter(centros[clust,0],centros[clust,1],centros[clust,2],marker='o',s=500,facecolor=centros[clust],edgecolors='k',zorder=500)
        axis.plot(np.zeros(50)+centros[clust,0], np.linspace(1,centros[clust,1]), np.zeros(50)+centros[clust,2],c='k', linewidth=1,linestyle='--',zorder=1)
        axis.plot(np.zeros(50)+centros[clust,0], np.zeros(50)+centros[clust,1], np.linspace(0,centros[clust,2]),c='k', linewidth=1,linestyle='--',zorder=1)
    axis.set_xlabel("Red",fontsize='x-large')
    axis.set_ylabel("Green",fontsize='x-large')
    axis.set_zlabel("Blue",fontsize='x-large')
    axis.set_xlim(min(r)-0.05,max(r)+0.05)
    axis.set_ylim(min(g)-0.05,max(g)+0.05)
    axis.set_zlim(min(b)-0.05,max(b)+0.05)
    axis.view_init(elev, azim)
    plt.show()
#########################################################################
def recreate_image(codebook, labels, w, h):
    """Recreate the (compressed) image from the code book & labels"""
    return codebook[labels].reshape(w, h, -1)
#########################################################################
def plot_quantized_img(img, centros, asignaciones,algo='K-Means'):
    k = len(centros)
    plt.figure(figsize=(20,10))
    plt.axis("off")
    plt.title(f"Quantized image ({k} colors, {algo})")
    cmap_name_km = 'km'
    cmap_km = LinearSegmentedColormap.from_list(cmap_name_km, centros[np.mean(centros[:,:],axis=1).argsort()], N=k)
    plt.imshow(recreate_image(centros, asignaciones, img.shape[0], img.shape[1]),cmap=cmap_km)
    plt.colorbar(ticks=[],fraction=0.0335, pad=0.01)
    plt.show()
#########################################################################
def plot_colorspace_con_asignaciones(img, centros, asignaciones, elev=35, azim=-35, algo='K-Means'):

    fig = plt.figure(figsize=(16,16))
    axis = fig.add_subplot(1, 1, 1, projection="3d",computed_zorder=False) # 3D plot with scalar values in each axis

    im = img[:,:]
    r, g, b = list(im[:,:,0].flatten()), list(im[:,:,1].flatten()), list(im[:,:,2].flatten())

    axis.scatter(r, g, b, facecolor=centros[asignaciones], marker="o",edgecolors='k',linewidth=0.02,alpha=0.3,zorder=2)
    for clust in range(centros.shape[0]):
        axis.scatter(centros[clust,0],centros[clust,1],centros[clust,2],marker='o',s=500,facecolor=centros[clust],edgecolors='k',zorder=500)
        axis.plot(np.zeros(50)+centros[clust,0], np.linspace(1,centros[clust,1]), np.zeros(50)+centros[clust,2],c='k', linewidth=1,linestyle='--',zorder=1)
        axis.plot(np.zeros(50)+centros[clust,0], np.zeros(50)+centros[clust,1], np.linspace(0,centros[clust,2]),c='k', linewidth=1,linestyle='--',zorder=1)
    plt.title(algo,fontsize='xx-large', y=0.85)
    axis.set_xlabel("Red",fontsize='x-large')
    axis.set_ylabel("Green",fontsize='x-large')
    axis.set_zlabel("Blue",fontsize='x-large')
    axis.set_xlim(min(r)-0.05,max(r)+0.05)
    axis.set_ylim(min(g)-0.05,max(g)+0.05)
    axis.set_zlim(min(b)-0.05,max(b)+0.05)
    axis.view_init(elev,azim)
    plt.show()
#########################################################################
def plot_cuantizaciones_iniciales(img, k=3):

    k = 3

    distorsion_rand = np.zeros(36)
    distorsion_mas = np.zeros(36)

    fig11 = plt.figure(figsize=(16, 6), constrained_layout=False)
    outer_grid = fig11.add_gridspec(1, 2, wspace=0.05, hspace=0.1)
    fig11.suptitle("Cuantización dada por los clusters inciales (variando semilla)", fontsize=20)

    axs_out = outer_grid.subplots()  # Create all subplots for the inner grid.

    axs_out[0].set_title('Inicialización Aleatoria',fontsize='xx-large')
    axs_out[1].set_title('Inicialización ++',fontsize='xx-large')

    axs_out[0].set(xticks=[], yticks=[])
    axs_out[1].set(xticks=[], yticks=[])

    # gridspec inside gridspec
    inner_grid = outer_grid[0].subgridspec(6, 6, wspace=0, hspace=0)
    axs = inner_grid.subplots()  # Create all subplots for the inner grid.
    for c, ax in enumerate(axs.flatten()):
        centros, asignaciones, distancias_train = cuantizacion_con_centros_iniciales(img.reshape(-1,3), k, init='random', semilla=c+48)
        ax.imshow(recreate_image(centros, asignaciones, img.shape[0], img.shape[1]))
        ax.set(xticks=[], yticks=[])

    # gridspec inside gridspec
    inner_grid = outer_grid[1].subgridspec(6, 6, wspace=0, hspace=0)
    axs = inner_grid.subplots()  # Create all subplots for the inner grid.
    for c, ax in enumerate(axs.flatten()):
        centros, asignaciones, distancias_train = cuantizacion_con_centros_iniciales(img.reshape(-1,3), k, init='masmas', semilla=c+48)
        ax.imshow(recreate_image(centros, asignaciones, img.shape[0], img.shape[1]))
        ax.set(xticks=[], yticks=[])


    #plt.savefig('grafica_init.jpg')
    plt.show()
    
#########################################################################
def plot_optim_k(img):
    plt.figure(figsize=(18,7))

    rango_k = np.arange(2,12)
    distorsion_k = np.zeros(10)

    for k in rango_k:
        centros, asignaciones, distancias = k_means_fuaa(img.reshape(-1,3), k=k, init='random', semilla=30, verbose=False)
        plt.subplot(1,10,k-1)
        plt.title(f'k={k}',fontsize='x-large')
        plt.axis('off')
        #plt.imshow(recreate_image(centros,asignaciones.reshape(img[:,:,0].shape)[78:129,78:129].flatten(),51,51))
        plt.imshow(recreate_image(centros, asignaciones, img.shape[0], img.shape[1]))
        distorsion_k[k-2] = distorsion_fuaa(distancias,asignaciones)
    plt.show()

    plt.figure(figsize=(16,5))
    plt.plot(rango_k,distorsion_k,marker='o')
    plt.fill_between(rango_k, distorsion_k, np.zeros(10), alpha=0.2)
    plt.xticks(rango_k,fontsize='large')
    plt.yticks(fontsize='large')
    plt.xlabel('k',fontsize='x-large')
    plt.ylabel('Distorsion', fontsize='x-large')
    plt.show()
#########################################################################
